/************************************************************************************************************
 **   GEMM - a generic mobility model  for generating realistic mobility scenarios                                          
 **   Copyright (C) 2003, 2004 Suprio Ray
 **   Copyright (C) 2003, 2004 University of British Columbia, Canada                             
 **                                                                           
 ** This program is free software; you can redistribute it and/or modify      
 ** it under the terms of the GNU General Public License as published by      
 ** the Free Software Foundation; either version 2 of the License, or         
 ** (at your option) any later version.                                       
 **                                                                           
 ** This program is distributed in the hope that it will be useful,           
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of           
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            
 ** GNU General Public License for more details.                              
 **                                                                           
 ** You should have received a copy of the GNU General Public License        
 ** along with this program; if not, write to the Free Software             
 ** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *************************************************************************************************************/

/**  GemmModel.java
 *  @description 
 *  @author  Suprio Ray
 *  @date  Sept. 03, 2003
 *  @version GEMM 1.0
 */
 
package ca.ubc.cs.gemm; 

import java.io.*; 
import java.util.*;

import  ca.ubc.cs.gemm.util.Trace; 
import  ca.ubc.cs.gemm.util.Util; 

import edu.bonn.cs.iv.bonnmotion.*;


public class GemmModel extends GemmScenario {
	
	private Trace trace;
	private Trace mTrace; 
	private double curTime;
	boolean _debug=false;
	boolean osc = false;
	
	private static final String MODEL_NAME = Util.REALISTIC_MODEL;
      
	private static double updateFrequency = 1.0;
   
 
	// different roles and related parameters
	protected double maxspeed, angleStdDev, speedStdDev=0;
 
	/** Parameters for the attractor field used to choose destinations. */
	 protected double[] aFieldParams = null;
	 protected	AttractionPoints aField = null;
	
	 /**  Parameters for roles. */
	 // protected double[] roleParams = null;
	 protected String roleParams; 
	 protected int numRoles = 0;
	 protected ArrayList roles = null;
	 
     /** Parameters for the activities */
     protected double[]  actParams = null;	
     protected ArrayList activities = null;

      /** Parameters for group mobilty */
	  protected boolean isValidGroupMobility = false;
	  protected double[] groupMobilityParams = null;
	  
	 protected boolean parseArg(String key, String value) {
	 	
		if (key.equals("model")) {
		   if (!value.equals(MODEL_NAME)) {
			   System.out.println("wrong model " + value);
			  System.exit(-1);
		   }
		   return true;
	   } else if (key.equals("debug")) {
	   	 	_debug=true;
	   	 	return true;
	   }   else if (key.equals("updateFrequency")) {
		   updateFrequency = Double.parseDouble(value);
		   return true;  
	   } else if (key.equals("maxspeed")) {
		   maxspeed = Double.parseDouble(value);
		   return true;
	   } else if (key.equals("angleStdDev")) {
		   angleStdDev = Double.parseDouble(value);
		   return true;   
	   }  else if (key.equals("speedStdDev")) {
		   speedStdDev = Double.parseDouble(value);
		   return true;	   
	   }  
	   else if (key.equals("randomSeed")) {
		   randomSeed = Long.parseLong(value);
		   return true;
	   } else if (key.equals("x")) {
		   return true;
	   } else if (key.equals("y")) {
		   return true;
	   } else if (key.equals("inputX")) {
		   x = Double.parseDouble(value);
		   return true;
	   } else if (key.equals("inputY")) {
		   y = Double.parseDouble(value);
		   return true;
	   } else if (key.equals("Osc")) {
				  osc = true;
				  return true;
	   
	   // create random attraction points     
	   } else if (key.equals("randomAttrPts")) {
	   	    System.out.println("Random attraction points....");
	   	  try {
	   	 
			aFieldParams = parseDoubleArray(value);
			if (aFieldParams != null) {
				  aField = new AttractionPoints(x, y);
				  for (int p = 0; p < aFieldParams.length; p += 3) {
						if (aFieldParams.length - p < 3) {
							System.out.println("warning: attraction field argument list has wrong number of elements!");
							System.exit(0);
				  		}	
				  		else {
				  		
				  			double j = aFieldParams[p];
				  			System.out.println("x" + x + "y" + y);
				 			for (double k=0;k<j;k++) {	
								aField.add(new Position(x * randomNextDouble(), y * randomNextDouble()), aFieldParams[p+1], aFieldParams[p+2], k);								
				 			}
				  		} 
				  }				
			}
	   	  } catch (Exception e) {
			System.out.println("Error: while parsing random acts params");
			e.printStackTrace();
		}	
	   		return true;
	   // create a number of activities with random attraction points to visit		
	   } else if (key.equals("randomActs")) {
		
		int actLen =0;		
		// note: -V should precede -randomActs if both are to be used
		if (activities == null) 
				activities = new ArrayList();
		else 
			  actLen = activities.size(); 
			  
		actParams = parseDoubleArray(value);
				
		System.out.println("Number of Random Acts: "+ actParams.length);
			
				try {
						for (int i=0; i< actParams.length; i=i+5) {
							
							for (double j=0;(double)j<actParams[i];j++) {
								activities.add(new Activity(actParams[i+1],actParams[i+2], actParams[i+3], actParams[i+4], j, actLen + j));
								
							} 		
						}	
				} catch (Exception e) {
						System.out.println("Error: while parsing activities params");
						e.printStackTrace();
				}	
	   		return true;
	   } else
			  return super.parseArg(key, value);

	}

	protected boolean parseArg(char key, String val) {
			
	   switch (key) {
		   	
			/* format: <% nodes>, <maxSpeed>, <minSpeed>, <role type>, (<activity1>, <activity2>,... ) */
			case 'R': 
			   	   roleParams = new String(val);
				   return true;
			/* params for attraction points */	
			case 'A':
				   aFieldParams = parseDoubleArray(val);
					
				    if (aFieldParams != null) {
					   aField = new AttractionPoints(x, y);
					   aField.Add(aFieldParams);
				    }
					return true;
					
			/* params for activities */ 	
			case 'V': 
				actParams = parseDoubleArray(val);
				activities = new ArrayList();
				try {
					for (int i=0; i<actParams.length; i=i+6) {
						activities.add(new Activity(actParams[i],actParams[i+1], actParams[i+2], actParams[i+3], actParams[i+4], actParams[i+5]));
						
					}	
				} catch (Exception e) {
					System.out.println("Error: while parsing activities params");
				}	
				return true;		
			/* params for group mobility */
			case 'G' :
			
				isValidGroupMobility = true;
				groupMobilityParams = parseDoubleArray(val);
				
				return true; 		 
			case 'u' :
			   updateFrequency = Double.parseDouble(val);
			   return true;
				
			case 'a' :
			   angleStdDev = Double.parseDouble(val);
			   return true;	
			   			
			case 's' :
			   speedStdDev = Double.parseDouble(val);
			   return true;	   				
			case 'm':
			   maxspeed = Double.parseDouble(val);
			   return true;
				
			case 'h' :
			   maxspeed = Double.parseDouble(val);
			   return true;	
			default :
			   return super.parseArg(key, val);
		   }
	   }

	   public GemmModel(
	      int nodes,
		  double x,
		  double y,
		  double duration,
		  double ignore,
		  long randomSeed,
		  double updateFrequency, 
		  double numStroller,
		  double numPlayer,
		  double numBiker ) {
		  super(nodes, x, y, duration, ignore, randomSeed);
		  this.updateFrequency = updateFrequency;
		 
	    		   generate();
	   }
	   
	   public GemmModel(String[] args) {
		   go(args);
	   }
	
	   public void go(String[] args) {
		    super.go(args);
		   try {
			   generate();
			  
		   }
		   catch (Exception e) {
			   System.out.println("Exception in generate() " + e + "\n The trace\n");
			   e.printStackTrace();
		   }
		
	   }

	  double xDir(Position p1, Position p2) {
	 	
		if (p1.distance(p2) < 10)
						return 0;
	  		
	  		if ( (p2.x - p1.x) > 0)
	  			return 1;
	  		else if ( (p2.x - p1.x) == 0)
				return 0;
			else 
				return -1;	
	  }
	  
	  double yDir(Position p1, Position p2) {
	  	 
	  		if (p1.distance(p2) < 10)
	  			return 0;
	  		
	  		if  ((p2.y - p1.y) > 0) 
	  			return 1;
	  		else if ((p2.y - p1.y) == 0 )
	  			return 	0;
	  		else 
	  			return -1;			
	  }
	  
	   public void generate()  {

		System.out.println("simulation area :" + x + " x" + y + "  duration:" + duration + "  number of nodes:" + node.length );
		System.out.println("number of activities :" + activities.size());
		   preGeneration();

			if (_debug)	{	 	
			    trace      = new Trace();
		   		mTrace = new Trace(Util.GROUP_MOBILITY_TRACE);
		    }
			
			int numNodeInfluencedByCollisionAvoidance =0;
			int numNodeInfluencedByVelocityMatching =0;
			int numNodeInfluencedByInertia=0;
			int numNodeInfluencedByGroupcentering=0;
			 	  
			int numNodeNOTInfluencedByCollisionAvoidance =0;
			int numNodeNOTInfluencedByVelocityMatching =0;
			int numNodeNOTInfluencedByInertia=0;
			int numNodeNOTInfluencedByGroupcentering=0;
						  
			  if (isValidGroupMobility) {
			  	
					double probCollisionAvoidance=0;
					double  probVelocityMatching=0;
					double probGroupCentering=0;		
					double probInertia = 0;
					
										
					probCollisionAvoidance= groupMobilityParams[0];
					numNodeInfluencedByCollisionAvoidance = (int) Math.floor (  node.length * probCollisionAvoidance);
					numNodeNOTInfluencedByCollisionAvoidance = node.length - numNodeInfluencedByCollisionAvoidance;
				
					probVelocityMatching =    groupMobilityParams[1];
					numNodeInfluencedByVelocityMatching = (int) 	Math.floor (  node.length *  probVelocityMatching);
					numNodeNOTInfluencedByVelocityMatching =   node.length  - numNodeInfluencedByVelocityMatching;
				
					probInertia = groupMobilityParams[2];
					numNodeInfluencedByInertia = (int) Math.floor (  node.length * probInertia);
					numNodeNOTInfluencedByInertia = node.length - numNodeInfluencedByInertia;
				
				
					probGroupCentering = groupMobilityParams[3];
					numNodeInfluencedByGroupcentering = (int) Math.floor (  node.length *  probGroupCentering);
					numNodeNOTInfluencedByGroupcentering =  node.length  - numNodeInfluencedByGroupcentering;
					
					
			  }
			   
		   try {
				String val = roleParams;
				trace.println("The role params " + val);
				StringTokenizer st= new StringTokenizer(val, ",");
				roles = new ArrayList();
			    
			  	while (st.hasMoreTokens()) {
					double percentOfTotalNodes = Double.parseDouble(st.nextToken());
					
					double speedMax = Double.parseDouble(st.nextToken());
					double speedMin = Double.parseDouble(st.nextToken());
					
					int   roleType =  Integer.parseInt(st.nextToken());
					String actStr=val.substring( val.indexOf('(')+1, val.indexOf(')') );
					val = val.substring(val.indexOf(')') + 1);
					StringTokenizer ast = new StringTokenizer(actStr, ",");
					ArrayList acts = new ArrayList();
					while (ast.hasMoreTokens()) {
						double act = Double.parseDouble(ast.nextToken());
						acts.add((Activity)getActivityByType(act));	
					}
					trace.println("% nodes:" + percentOfTotalNodes + " speedmax:" + speedMax + " speedmin" + speedMin + " roleType:" + roleType + " acts :" + actStr);
					roles.add(new Role(percentOfTotalNodes,speedMax,speedMin,roleType, acts));
				 	 		
					numRoles++;	
					if (_debug)	
						trace.println("Total number of roles " + numRoles + "the remaining string " + val);
					st= new StringTokenizer(val, ",");
					}
					
		  } catch (Exception e) { 
			if (_debug)	
					trace.println("Error: while parsing Role parameters\n"); e.printStackTrace();
		  }    
               
		   double maxX = x;
		   double maxY = y;
		   double minX = 0;
		   double minY = 0;
		
		   trace.println(" x="+ x+ " y="+ y);
		   double []   rolesPerNum = new double [numRoles];
		   double checkTot =0;
		   for (int i=0; i< numRoles; i++) {
		   	
				rolesPerNum[i] =  ((Role) roles.get(i)).getPer() * node.length/ 100;
				if (_debug)	
					trace.println("role " + ((Role) roles.get(i)).getRoleType() + " : num nodes  " + rolesPerNum[i]);
		   		checkTot += rolesPerNum[i];
		   }
			if (checkTot != node.length) {
				System.out.println ("Error: % roles " + checkTot + " - don't sum upto 100");
				System.exit(1);
			
			}
			
			Random rndRole = new Random();
			Random rndGrpMobility = new Random();
			
			// double speed, dir;
			Position src = null;
			for (int i=0; i< node.length; i++) {
				
					node[i]= new GemmMobileNode();
					boolean isRoleSet = false;
					do {
					
						int rlRnd=rndRole.nextInt(numRoles);
					
						if (rolesPerNum[rlRnd] != 0)
						{
							rolesPerNum[rlRnd] = rolesPerNum[rlRnd] - 1; 
							node[i].setRole (new Role(getRoleByType((double)rlRnd)));
							node[i].setActive(false);
							isRoleSet = true;	
							
							src = new Position(x * randomNextDouble(), y * randomNextDouble());
							if (!node[i].add(0.0, src)) {
								if (_debug)	
									trace.println("RealisticModel.init: error while adding node movement (1)");
								System.exit(0);
							}

							node[i].setCurPos(src);
							double dir = randomNextDouble() * 2 * Math.PI;
							
		
							double maxspeed =  node[i].getRole().getMaxSpeed();
							double minspeed =  node[i].getRole().getMinSpeed();
							
							// Random Waypoint speed model: double speed = (maxspeed - minspeed) * randomNextDouble() + minspeed; 
							if (maxspeed == 0) maxspeed = 0.00000001;
							if (minspeed == 0) minspeed = 0.00000001;
	                        // several ways of generating random speed:
							//double betaRnd = randomNextBeta(maxspeed, minspeed);
							// double speed = (maxspeed - minspeed) * betaRnd + minspeed;
							//double speed = (maxspeed - minspeed) * randomNextGaussian() + minspeed;
							double speed = (maxspeed - minspeed) * randomNextDouble() + minspeed;
							
							
							if (speed <  minspeed)
								speed = minspeed;
	
							node[i].setDir(dir);
							node[i].setSpeed(speed);
							
							// randomize the last perfomed action time
							ArrayList actvts = node[i].getRole().getActivities();
							for (int a=0;a<actvts.size();a++)
								node[i].getRole().setLastTimePerformed(((Activity)actvts.get(a)).getActivityType() ,randomNextDouble()% duration);   		
							
						}
						
					  } while (isRoleSet == false); // end setting the role
					  
					  
					if (isValidGroupMobility) {
						
						// collision avoidance
						isRoleSet = false;
						do {
								int  nca= rndGrpMobility.nextInt(2);
								if ( (nca == 1) && (numNodeInfluencedByCollisionAvoidance != 0)) {
									isRoleSet =true;				
									numNodeInfluencedByCollisionAvoidance--;	
									node[i].CollisionAvoidance = true;		
								}
								else if ((nca == 0) && (numNodeNOTInfluencedByCollisionAvoidance != 0)) {
									isRoleSet =true;	
									numNodeNOTInfluencedByCollisionAvoidance--;
									node[i].CollisionAvoidance = false;
								}
								 
						} while (isRoleSet == false);
					
					    // Group Centering
						isRoleSet = false;
						do {
									int  nca= rndGrpMobility.nextInt(2);
									if ( (nca == 1) && (numNodeInfluencedByGroupcentering != 0)) {
										isRoleSet =true;		
										numNodeInfluencedByGroupcentering--;	
										node[i].Groupcentering = true;				
									}
									else if ((nca ==0) && (numNodeNOTInfluencedByGroupcentering != 0)) {
										isRoleSet =true;	
										numNodeNOTInfluencedByGroupcentering--;	
										node[i].Groupcentering = false;
									}
								 
						} while (isRoleSet == false);
					
		                  // Velocity Matching 		
							isRoleSet = false;
							do {
									int  nca= rndGrpMobility.nextInt(2);
									if ( (nca == 1) && (numNodeInfluencedByVelocityMatching != 0)) {
											isRoleSet =true;		
											numNodeInfluencedByVelocityMatching--;	
											node[i].VelocityMatching = true;
															
									}
									else if ((nca ==0) && (numNodeNOTInfluencedByVelocityMatching != 0)) {
											isRoleSet =true;
											numNodeNOTInfluencedByVelocityMatching--;
											node[i].VelocityMatching = false;	
									}
								 
								} while (isRoleSet == false);
								
						// Inertia 
						isRoleSet = false;
						do {
							int  nca= rndGrpMobility.nextInt(2);
							if ( (nca == 1) && (numNodeInfluencedByInertia != 0)) {
									isRoleSet =true;		
								    numNodeInfluencedByInertia--;	
									node[i].Inertia = true;
															
							}
							else if ((nca ==0) && (numNodeNOTInfluencedByInertia != 0)) {
									isRoleSet =true;
									numNodeNOTInfluencedByInertia--;
									node[i].Inertia = false;	
							}
								 
						} while (isRoleSet == false);
				}						
								 					
													
			} //		end for:  all node's initial position

		 
			  { double avgSpd =0;
			  
			   for (int i=0; i< node.length; i++) {
				   double speed = node[i].getSpeed();
				  // System.out.println("***speed " + speed);
				   avgSpd = avgSpd + speed;
			   }
			   avgSpd = avgSpd/node.length;
		
			   System.out.println("avg dur " + (Dur/Cnt) + "  Count " + Cnt);
			   System.out.println("avg speed " + avgSpd + "  Count " + sCnt);
			  }
		 
		      double avgDist=0;
		      int dstCnt =0;
			 
			   
				for (int i=0; i< node.length; i++) {
					Role rl = node[i].getRole();
					ArrayList osciTest =  ((ArrayList) rl.getActivities()); 
				  				
					//osci: if this is an oscilating activity: first entry must be -1 (Osc)
					if ( ((Activity)osciTest.get(0)).getActivityType() == -1 ) {
								node[i].osci = true;
								node[i].origPos = new Position( node[i].getCurPos());
								node[i].osciAct = new Activity((Activity)osciTest.get(0));
								osciTest.remove(0);
								
								node[i].getRole().setActivities(osciTest);
					}
				}	
				 if (_debug)	
				 	trace.println("The duration of the simulation " + duration);	
				 for (double t=0.0+ updateFrequency; t< duration; t=t+updateFrequency) {
				 	
					setCurTime(t);
				 	
				 	for (int i=0; i< node.length; i++) {
						Role rl = node[i].getRole();  
						node[i].setCurTime(t);
						
				  		Position dst = null;
				  		// Check current activity
				  		if (node[i].isActive() == false) {
				  			
				  			// check for the activities he can take part in
				  			// choose the most immediate activitiy	
				  				
						    	Object [] acts=  ((ArrayList) rl.getActivities()).toArray();
				  								  				
				  				double  minTrigTime = duration;
								Activity curActivity = null;
				  				for (int j=0;j<acts.length;j++) {
				  					Activity act = (Activity)acts[j];
				  					double triggerTime = calculateNextTriggerTime(act, node[i].getRole());
				  					
				  					if (triggerTime < minTrigTime ) {
				  						minTrigTime = triggerTime;
				  						curActivity = act;
				  					}
				  				}	
								if (_debug)	
				  					trace.println("#### At time " + t + " Node " + i+ "choses activity " + curActivity.getActivityType() );
				  
				  				node[i].setCurActivity(curActivity);		
				  				double aptype = curActivity.getAttractionPointToGo();
				  				
								if ( aptype == Activity.NO_PARTICULAR_ATTRACTION_POINT) {
				  					
									node[i].setActive(true);	
									node[i].setNoTarget();  
									//set the activity duration
									double adur= calculateActivityDuration(curActivity);
									node[i].setActivityDuration(adur);					
								}
				  				else if ( aptype == Activity.RANDOM_DIRECTION) {  // random attraction point
									node[i].setActive(true);
									// choose a random attraction point
									double xPos = x * randomNextDouble();
									double yPos = y * randomNextDouble();
									
									Position tDst = new Position(xPos,yPos);
									
									node[i].setTargetDst(tDst);
									
									Position cPos= node[i].getCurPos();
									
									double speed = node[i].getSpeed();
								
				    				double xdir = xDir(cPos, tDst);
				    				double ydir = yDir(cPos, tDst);
									double dir = Angle(cPos, tDst);
									
									speed = getNewSpeed(speed,rl);	
									
									if (speed > 0.0) {
										 dst =  new Position(
										   cPos.x +  xdir * Math.cos(dir) * updateFrequency * speed,
									       cPos.y + ydir * Math.sin(dir) * updateFrequency * speed
										);
																
				      
										if (dst.reaches(tDst)) {
											   node[i].setTargetReached();
											   if (_debug)	
												trace.println("At time:" + t + " Node:" + i + " of Role type:"+ node[i].getRole().getRoleType() + "Reaches destination (" + tDst.x + ","+ tDst.y + ")" );
										    	//set the activity duration
												double adur= calculateActivityDuration(node[i].getCurActivity());
												node[i].setActivityDuration(adur);
										}
										else {
											
											// Perception and Group Mobility behavior
											if (isValidGroupMobility) {
												GroupMobility gm = determineGroupMobilityBehavior(i, speed,  dir,  xdir,  ydir,  dst );
												dst = gm.p;
												dir = gm.dir;
												speed = gm.speed;						
											} 		
											
										}   // end group mobility

										if (!node[i].add(t, dst)) {
											   System.err.println("init: error while adding node movement (2)");
											   System.exit(0);
										}
										node[i].setCurPos(dst); 
										if (_debug)	
											trace.println("At time:" + t + " Node:" + i + " of Role type:"+ node[i].getRole().getRoleType() + "is at: (" + dst.x + ","+ dst.y + ") -> going to attraction point (" + tDst.x + ","+ tDst.y + ") whose distance=" + tDst.distance(dst) +  "speed " + speed );
									   
								   }
								   
								   node[i].setDir(dir);
								   node[i].setSpeed(speed);	
				  				}
				  				else {
									// set the target
									node[i].setActive(true);
									
									Position tDst = aField.getPosByType(aptype);
									if (tDst != null) {
									
										node[i].setTargetDst(tDst);					
										Position cPos= node[i].getCurPos();
				
										if (node[i].dst ==0) { 
											node[i].dst = cPos.distance(tDst); 
											avgDist += cPos.distance(tDst);
											dstCnt++;	
										}							
									
										double speed = node[i].getSpeed();
										speed = getNewSpeed(speed,rl);
										
										double xdir = xDir(cPos, tDst);
										double ydir = yDir(cPos, tDst);
										double dir = Angle(cPos, tDst);
		
									
										if (speed > 0.0) {
										 dst =  new Position(
										 cPos.x +  xdir * Math.cos(dir) * updateFrequency * speed,
										  cPos.y + ydir * Math.sin(dir) * updateFrequency * speed
										 );
										 
				      
											if (dst.reaches(tDst)) {
												   node[i].setTargetReached();
													trace.println("At time:" + t + " Node:" + i + " of Role type:"+ node[i].getRole().getRoleType() + "Reaches destination (" + tDst.x + ","+ tDst.y + ")" );
													//set the activity duration
													double adur= calculateActivityDuration(node[i].getCurActivity());
													node[i].setActivityDuration(adur);
											}
											else {
												// Perception and Group Mobility behavior
												if (isValidGroupMobility) {
													GroupMobility gm = determineGroupMobilityBehavior(i, speed,  dir,  xdir,  ydir,  dst );
													dst = gm.p;
													dir = gm.dir;
													speed = gm.speed;						
												} 		
											 }   

											 if (!node[i].add(t, dst)) {
												   System.out.println("init: error while adding node movement (2)");
												   System.exit(0);
											  }
											  node[i].setCurPos(dst); 
											  trace.println("At time:" + t + " Node:" + i + " of Role type:"+ node[i].getRole().getRoleType() + "is at: (" + dst.x + ","+ dst.y + ") -> going to attraction point (" + tDst.x + ","+ tDst.y + ") whose distance=" + tDst.distance(dst) +  "speed " + speed );
									   }
									   // node[i].setDir(dir);
									   node[i].setSpeed(speed);		
									}
									else {
										System.out.println("No attr found with type " + aptype);
										System.exit(0);
									}
				  				}
								
				  		} // end if node inactive	
				  		else {
				  			// node already active
				  			
				  			// in motion? 
				  			// if (node[i].isInMotion()) {

							
							if ( (node[i].isTargetSet()) &&  (!node[i].isTargetReached()))  {   // target already set and we're heading toward that
 
								Position cPos= node[i].getCurPos();
								Position tDst = node[i].getTargetDst();
				  
								
								double speed = node[i].getSpeed();
								speed = getNewSpeed(speed,rl);
										
								double xdir = xDir(cPos, tDst);
								double ydir = yDir(cPos, tDst);
								double dir = Angle(cPos, tDst);
		
								if (speed > 0.0) {
									 dst =  new Position(
									 cPos.x +  xdir * Math.cos(dir) * updateFrequency * speed,
									 cPos.y + ydir * Math.sin(dir) * updateFrequency * speed
									 );
									 
				      
									if (dst.reaches(tDst)) {
										node[i].setTargetReached();
										if (_debug)	
											trace.println("Node " + i+ " reached the destination " + tDst.toString());
										// set the activity duration
										double adur= calculateActivityDuration(node[i].getCurActivity());
										node[i].setActivityDuration(adur);
									}
									else {
											// Perception and Group Mobility behavior
											if (isValidGroupMobility) {
												GroupMobility gm = determineGroupMobilityBehavior(i, speed,  dir,  xdir,  ydir,  dst );
												dst = gm.p;
												dir = gm.dir;
												speed = gm.speed;						
											} 		
									}   

									if (!node[i].add(t, dst)) {
										System.out.println(
											   "Error while adding node movement (2)");
										System.exit(0);
									}
									node[i].setCurPos(dst);
									if (_debug)	
										trace.println("At time:" + t + " Node:" + i + " of Role type:"+ node[i].getRole().getRoleType() + "is at: (" + dst.x + ","+ dst.y + ") -> going to attraction point (" + tDst.x + ","+ tDst.y + ") whose distance=" + tDst.distance(dst) +  "speed " + speed );
								}
								node[i].setDir(dir);
								node[i].setSpeed(speed);
				  
							}
							else if ( (node[i].isTargetSet()) &&  (node[i].isTargetReached()))  {   // target already reached and we're doing the activity at an attraction point
								
									
								double aDur = node[i].getActivityDuration();
				
								aDur--;
								if (aDur <= 0) {
									
									// test for osci
									if (node[i].osci == true) {
									
										// if this is the origPos, then setActive false
										if (node[i].getTargetDst().equals(node[i].origPos)) {
											node[i].setActive(false);  
											node[i].setNoTarget();
											  
										}
										else {
											// choose the orig pos as the dest
											node[i].getRole().setLastTimePerformed(node[i].getCurActivity().getActivityType() ,node[i].getCurTime());
											
										    node[i].setActive(true);  
										    node[i].setTargetDst(node[i].origPos); 
									 	    node[i].setCurActivity(node[i].osciAct);
											double adur= calculateActivityDuration(node[i].getCurActivity());
											node[i].setActivityDuration(adur);	
									
										}
									}
									else {
									node[i].setActive(false);  
									node[i].setNoTarget();  
									node[i].getRole().setLastTimePerformed(node[i].getCurActivity().getActivityType() ,node[i].getCurTime());
									}   		
								}
								else {		
									node[i].setActivityDuration(aDur);							
								}
							} // end going to attraction point
							
				  			else if (! (node[i].isTargetSet())) { // means activity with no attraction point to go to eg. strolling
				  				
								double aDur = node[i].getActivityDuration();
								trace.println("start: activity duration " + aDur + " for node " + i + "with no target destination");
								aDur--;
								if (aDur <= 0.0) {
									trace.println("over: activity duration " + aDur + " for node " + i + "with no target destination");
									node[i].setActive(false);
									node[i].getRole().setLastTimePerformed(node[i].getCurActivity().getActivityType() ,node[i].getCurTime());   	
								}
								else {		
										node[i].setActivityDuration(aDur);
											
										Position cPos= node[i].getCurPos();
			
										double speed = node[i].getSpeed();
										speed = getNewSpeed(speed,node[i].getRole());
				   
										double dir= node[i].getDir();
										if (speed > 0.0) {
											
											dir = getNewDir(dir, cPos);
											dst = new Position(
													 cPos.x + Math.cos(dir) * updateFrequency * speed,
													 cPos.y + Math.sin(dir) * updateFrequency * speed
													 );
											if (!node[i].add(t, dst)) {
												System.out.println("Error while adding node movement (2)");
												System.exit(0);
											}
											node[i].setCurPos(dst);	
											if (_debug)	
												trace.println("At time:" + t + " Node:" + i + " of Role type:"+ node[i].getRole().getRoleType() + "is at: (" + dst.x + ","+ dst.y + ")");	
										}
									node[i].setDir(dir);
									node[i].setSpeed(speed);
				  			}			
											  			
				  		}  // end: motion with no particular direction
				  		 
				 	} // end node active
				 
				 } // end for each node
			 } // end for update frequency

		
		
		postGeneration();
		System.out.println("End of generation...");
		
		
		double avgSpd =0;
		for (int i=0; i< node.length; i++) {
			double speed = node[i].getSpeed();
			avgSpd = avgSpd + speed;
		}
		avgSpd = avgSpd/node.length;
		
		System.out.println("\navg dur " + (Dur/Cnt) + "  Count " + Cnt);
		System.out.println("avg speed " + avgSpd + "  Count " + sCnt);
		System.out.println("avg distance" + (avgDist/dstCnt) + "  count" + dstCnt);
		
		
	   } // end generate
	 

   // Group Mobility behavior 
	class GroupMobility {
			Position p;
			double speed;
			double dir;
			GroupMobility(Position P, double Speed, double Dir) {
				p= P;
				speed = Speed;
				dir = Dir;
			}		
		}
		
	
	
	public  GroupMobility determineGroupMobilityBehavior(int i, double speed, double dir, double xdir, double ydir,  Position dst ) {
	
		
		node[i].nearbyNbrList = null;		
		node[i].grpNbrList = null;										
		HashSet nbr = new HashSet();
												
			// Perception			
			// All other nodes' locations are checked to see if any of them are within the sensitive zone for this node 		
			// Find out the neighbor nodes:
			for (int j=0;j<node.length;j++) {
				if ((i != j) && ( node[j].getCurPos().distance(dst) <= GemmMobileNode.SENSITIVITY_ZONE )) {
					nbr.add( node[j].getCurPos());  // insert in the neighbor position hashset
					if (node[i].nearbyNbrList == null )  node[i].nearbyNbrList = new ArrayList();
						node[i].nearbyNbrList.add(new Integer(j));
				}
				
				// group centering				
				if ( (i != j) && ( node[j].getCurPos().distance(dst) <= GemmMobileNode.GROUP_RANGE )
				      && ( node[j].getTargetDst() != null) 
				      && (node[i].getRole().getMaxSpeed() == node[j].getRole().getMaxSpeed())
				      && (node[i].getRole().getMinSpeed() == node[j].getRole().getMinSpeed())		
				) {
						if (node[j].getTargetDst().equals(node[i].getTargetDst())) {			
							if (node[i].grpNbrList == null )  node[i].grpNbrList = new ArrayList();
																node[i].grpNbrList.add(new Integer(j));
						}				
				}				
								  		
			
			}
												  	
		if (node[i].VelocityMatching) {
			// Velocity matching 
			// check for nearby 3/4 nodes that are roughly moving toward the same direction and  
			// and take the average of the velocities of nearby nodes and 
			// set that as its velocity,
			double avgVel=0;
			int count=0;
			if (node[i].nearbyNbrList != null) {
					for (int k=0;k<node[i].nearbyNbrList.size();k++)  {
								if (  (node[i].getRole().getMaxSpeed() == node[((Integer)node[i].nearbyNbrList.get(k)).intValue()].getRole().getMaxSpeed())
										&& (node[i].getRole().getMinSpeed() == node[((Integer)node[i].nearbyNbrList.get(k)).intValue()].getRole().getMinSpeed())
								    ) {
												avgVel+= node[ ((Integer)node[i].nearbyNbrList.get(k)).intValue()].getSpeed();
												count++;
								}		
					}		
				if (count > 0 ) { 								
					avgVel= avgVel/	count;
					if (node[i].Inertia) {
							//	if the difference b/w that and corrent velocity is not more than INERTIA_BOUND
							 // otherwise increase/decrease by INTERTIA_BOUND 
							if ( Math.abs(node[i].getSpeed() - avgVel) >  GemmMobileNode.INERTIA_BOUND) {
									if (node[i].getSpeed() > avgVel)
											speed = speed - GemmMobileNode.INERTIA_BOUND;
									else 
											speed = speed +  GemmMobileNode.INERTIA_BOUND;
							}	
							else 
									speed = avgVel;
						}
						else {
									speed = avgVel;
						}	
			        }	
				}
													
		}
			//		Group centering-velocity matching
			 double avgGrpVel=0;
			 if (node[i].grpNbrList != null) {
			 for (int k=0;k<node[i].grpNbrList.size();k++)
					 avgGrpVel+= node[ ((Integer)node[i].grpNbrList.get(k)).intValue()].getSpeed();
				
			 avgGrpVel= avgGrpVel /	node[i].grpNbrList.size();		
			 speed = avgGrpVel;			
																							  					  	 
			 }	
			 if (_debug)	
		     	mTrace.println("The  velocity" + speed + "   of node" + i);	
												  	
			// Collision avoidance
			// If any of the nodes are within the sensitive_zone, avoid collision by choosing a new destination
			if ( (node[i].CollisionAvoidance) && (node[i].nearbyNbrList != null) ) {
				boolean collided = true;
				Random rPon = new Random(System.currentTimeMillis());
				while (collided) {
												  			
					/*double offsetX = randomNextDouble() * RMMobileNode.SENSITIVITY_ZONE;
					double offsetY = randomNextDouble() * RMMobileNode.SENSITIVITY_ZONE;*/
															
					int pOn = rPon.nextInt(3) -1;
					double newDir = dir + pOn * randomNextDouble() * (Math.PI/4);
															
					dst =  new Position(
								node[i].getCurPos().x +  xdir * Math.cos(newDir) * updateFrequency * speed, 
								node[i].getCurPos().y + ydir * Math.sin(newDir) * updateFrequency * speed
						   );
												  				
					if (!nbr.contains(dst)) {
						if (_debug)	
							mTrace.println(i + " Chosen to avoid collision:  at time  " + node[i].getCurTime()+ "   - dirDiff" + (newDir - dir));
						collided = false;
						dir = newDir;
						
					}
				}
			}
					
							  
		return new GroupMobility(dst, speed, dir);
	}
	 
public Activity getActivityByType(double type) {
	for (int i=0;i<activities.size();i++) {
		if ( ((Activity)activities.get(i)).matches(type) )
			return 	(Activity)activities.get(i);
	}
	return null;
}


public Role getRoleByType(double type) {
		
		for (int i=0;i<roles.size();i++)
		{
				if ( ((Role)roles.get(i)).getRoleType() == type)
					return (Role)roles.get(i);	 
		}
		System.out.println("Error: No role found by the type" + type);
		System.exit(1);
		return null;
}

public double calculateNextTriggerTime(Activity act, Role role) {
	
	double maxT = act.getMaxTriggerTime();
	double minT = act.getMinTriggerTime();
	
	if (maxT <= 0.0)
		maxT = 0.1;
	if (minT <= 0.0)
		minT = 0.1;
			
	double betaRnd = randomNextBeta(maxT, minT);	
	double triggerTime = (maxT - minT ) * betaRnd + minT - ( getCurTime() - role.getLastTimePerformed(act.getActivityType()));

	return triggerTime;	
}

static double Dur=0;
static int Cnt =0;

public double calculateActivityDuration(Activity act) {
	
	double maxT = act.getMaxActDuration();
	double minT = act.getMinActDuration();
	
		if (maxT <= 0.0)
			maxT = 0.000000001;
		if (minT <= 0.0)
			minT = 0.000000001;
			
	double betaRnd = randomNextBeta(maxT,minT);
	

	double dur = act.getMaxActDuration() * randomNextDouble();
	dur = Math.ceil(dur);
	
	if (dur < act.getMinActDuration())
			dur = act.getMinActDuration();
			
			Dur = Dur + dur;
			Cnt++;

	return dur;
}

double getCurTime() {
	return curTime;
}

void setCurTime(double t) {	
	curTime =t;
}

double Angle (Position p1, Position p2) {
   if (!p1.reaches(p2)) {
 	
 	
	   double x = ((p2.x - p1.x) >= 0) ? (p2.x - p1.x) : ((-1) * (p2.x - p1.x));
	   double y = ((p2.y - p1.y) >= 0) ? (p2.y - p1.y) : ((-1) * (p2.y - p1.y));  
	
	
	   if (x == 0) 
		   return Math.PI/2;
	   else {	
 	
		   double angle = Math.atan(y/x) ;
		  
		   return angle;
	   } 	
   }
   else 
	   return 0;
}


public double angle (Position p1, Position p2) {


	double dir = Position.angle(p1,p2) ;

	if 	( ( (p1.x - p2.x) >= 0)	&& ((p1.y - p2.y ) >= 0) ) {
		if (dir <= Math.PI /2 )
			dir =  Math.PI + dir;
		else 
			dir = 2 * Math.PI - dir;	
	}
	else if ( ( (p1.x - p2.x) < 0)	&& ((p1.y - p2.y ) >= 0) ) {
		if (dir <= Math.PI/2 )
			dir = 2 * Math.PI - dir;
		else 
			dir = Math.PI + dir;	
	}	
	else if ( ( (p1.x - p2.x) < 0)	&& ((p1.y - p2.y ) < 0) ) {
		if (dir <= Math.PI /2)
			;
		else 
			dir = Math.PI - dir;
	}
	else if ( ( (p1.x - p2.x) >= 0)	&& ((p1.y - p2.y ) < 0) ) {
			
		if (dir <= Math.PI /2 )
			dir = Math.PI - dir;
		else 
			;	//  :-)
	}
	
		return dir;
}
	

	public double getNewDir(double oldDir, Position pos) {
	// move away from the border in case we are getting too close
	if (pos.x < 0)
		if (pos.y < 0)
			oldDir = 0.25 * Math.PI;
		else if (pos.y > y)
			oldDir = 1.75 * Math.PI;
		else
			oldDir = 0.0;
	else if (pos.x > x)
		if (pos.y < 0)
			oldDir = 0.75 * Math.PI;
		else if (pos.y > y)
			oldDir = 1.25 * Math.PI;
		else
			oldDir = Math.PI;
	else if (pos.y < 0)
		oldDir = 0.5 * Math.PI;
	else if (pos.y > y)
		oldDir = 1.5 * Math.PI;

	return randomNextGaussian() * angleStdDev + oldDir;
}


public double getNewSpeed(double oldSpeed, Role role) {
	
	/* Gauss-Markov speed model 
	double maxSpeed = role.getMaxSpeed();
	
	double speed = oldSpeed + randomNextGaussian() * speedStdDev;
	if (speed < role.getMinSpeed())
			speed = role.getMinSpeed();
	else if (speed > maxSpeed) {
		speed = maxSpeed;
	}
	*/
	
	/* Random Waypoint speed model: double speed = (maxspeed - minspeed) * randomNextDouble() + minspeed; */
	double maxspeed = role.getMaxSpeed();
	double minspeed = role.getMinSpeed();
	
	
	//	beta distribution
	/*
	if (maxspeed == 0) maxspeed = 0.00000001;
	if (minspeed == 0) minspeed = 0.00000001;
	
	double betaRnd = randomNextBeta(maxspeed, minspeed);
	double speed = oldSpeed + betaRnd * speedStdDev;
	*/
	
	// Gaussian distribution
	double speed = oldSpeed + randomNextGaussian() * speedStdDev;
	
	if (speed > maxspeed)
		speed = maxspeed;
	if (speed < minspeed)
		speed = minspeed;	
		
	Speed = Speed + speed;
	sCnt++;	
			
	return speed;
}
	
static double Speed =0; 
static int  sCnt =0;


public void write(String _name) throws FileNotFoundException, IOException {
	String[] p = new String[5];

	p[0] = "model=" + MODEL_NAME;
	p[1] = "updateFrequency=" + updateFrequency;
	p[2] = "maxspeed=" + maxspeed;
	p[3] = "angleStdDev=" + angleStdDev;
    p[4] = "speedStdDev=" + speedStdDev;

	super.write(_name, p);
}

public static void printHelp() {
	Scenario.printHelp();
	System.out.println( MODEL_NAME + ":" );
	System.out.println("\t-a <angle standard deviation>");
	System.out.println("\t-h <max. speed>");
	System.out.println("\t-q <speed, angle update frequency>");
	System.out.println("\t-s <speed standard deviation>");
}

 
} //end 
