/************************************************************************************************************
 **   GEMM - a generic mobility model  for generating realistic mobility scenarios                                          
 **   Copyright (C) 2003, 2004 Suprio Ray
 **   Copyright (C) 2003, 2004 University of British Columbia, Canada                             
 **                                                                           
 ** This program is free software; you can redistribute it and/or modify      
 ** it under the terms of the GNU General Public License as published by      
 ** the Free Software Foundation; either version 2 of the License, or         
 ** (at your option) any later version.                                       
 **                                                                           
 ** This program is distributed in the hope that it will be useful,           
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of           
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            
 ** GNU General Public License for more details.                              
 **                                                                           
 ** You should have received a copy of the GNU General Public License        
 ** along with this program; if not, write to the Free Software             
 ** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *************************************************************************************************************/

/**  Role.java
 *  @description 
 *  @author  Suprio Ray
 *  @date  Sept. 14, 2003
 *  @version GEMM 1.0
 */
package ca.ubc.cs.gemm; 

import java.util.*;


public class Role {
	
	double percentOfTotalNodes;
	double speedMax;
	double speedMin;
	double roleType;
	ArrayList activities;	
	
	Role(double potn, double sM, double sm, double rT, ArrayList acts)	{
		percentOfTotalNodes=potn;
		speedMax = sM;
		speedMin = sm;
		roleType = rT;
		// activities = acts;
		activities = new ArrayList();
		
		for (int i=0;i<acts.size();i++)
					activities.add(new Activity((Activity)acts.get(i)));	
	} 	
	Role(Role r) {
		this.percentOfTotalNodes = r.percentOfTotalNodes;
		this.speedMax = r.speedMax;
		this.speedMin = r.speedMin;
		this.roleType = r.roleType;
		activities = new ArrayList();
		for (int i=0;i<r.activities.size();i++)
			this.activities.add(new Activity((Activity)r.activities.get(i)));	
	}
	
	public double getPer() {
		return 	percentOfTotalNodes;
	} 
	public double getMaxSpeed() {
		return speedMax;
	}
	public double getMinSpeed() {
			return speedMin;
	}
	
	public double getRoleType() {
		return roleType;
	}

	public ArrayList getActivities() {
			return activities;
	}
	// osci
	public void setActivities(ArrayList acts) {
		activities = new ArrayList();
		
		for (int i=0;i<acts.size();i++)
					activities.add(new Activity((Activity)acts.get(i)));	
	}
	 
	public Activity getActivityByType(double type) {
		for (int i=0;i<activities.size();i++) {
			if ( ((Activity)(activities.get(i))).matches(type) )
				return 	(Activity)activities.get(i);
		}
		return null;
	}
	
	public double getLastTimePerformed(double actType) {
			Activity act = getActivityByType(actType);
			return act.getLastTimePerformed();
	}
	public void setLastTimePerformed(double actType, double time) {
		Activity act = getActivityByType(actType);
		act.setLastTimePerformed(time);
	}
}
